#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Import GitHub bugs into a Trac database.

Requires:  Trac 1.2 or newer from http://trac.edgewall.com/ (tested with 1.5)
           Python 3 from http://www.python.org/

Example use:
  python github2trac.py --tracenv /usr/local/trac-projects/myproj/ \
    --clean --append

Based on mantis2trac version 1.6, itself based on bugzillat2trac version 1.6

Authors:
- Adrien Destugues (pulkomandy@pulkomandy.tk) - GitHub support, Python 3 update, Trac 1.4 update
- Steffen Mecke (stm2@users.sourceforge.net)
- Matthew Parmelee (mparmelee@interworx.com)
- John Lichovník (licho@ufo.cz)
- Anton Stroganov (stroganov.a@gmail.com)
- Joao Prado Maia (jpm@pessoal.org)
- Paul Baranowski (paul@paulbaranowski.org)
- Dmitry Yusupov <dmitry_yus@yahoo.com> - bugzilla2trac.py
- Mark Rowe <mrowe@bluewire.net.nz> - original TracDatabase class
- Bill Soudan <bill@soudan.net> - Many enhancements 

Notes:
  - Releases should be imported as versions in Trac, this is not done yet
  - Some fields may not be imported properly or at all
  - Mantis2Trac used direct communication with the database (SQL queries) for things that now have
    a higher level Python API, code should be converted to that API instead
  - Importing attachments is not implemented (couldnt get it to work, 
    and we didnt have enough attachments to justify spending time on this)
    "Clean" will not delete your existing attachments.  There is code in here
    to support adding attachments, but you will have to play with it to 
    make it work.  If you search for the word "attachment" you will find
    all the code related to this.
"""
from urllib.parse import quote
from datetime import datetime, date

import json
import time
import hashlib
import requests

###
### Conversion Settings -- edit these before running if desired
###

# Path to the Trac environment.
TRAC_ENV = '/var/www/trac/projectname/'

# If true, all existing Trac tickets will be removed 
# prior to import.
TRAC_CLEAN = False
# If TRAC_CLEAN is true and this is true, tickets will be appended
TRAC_APPEND = True

# Enclose imported ticket description and comments in a {{{ }}} 
# preformat block?  This formats the text in a fixed-point font.
PREFORMAT_COMMENTS = False

# Ticket changes in Trac have the restriction where the
# bug ID, field, and time must be unique for all entries in the ticket 
# changes table.
# Mantis, for unknown reasons, has fields that can change two states 
# in under a second (e.g. "milestone":""->"1.0", "milestone":"1.0"->"2.0").
# Setting this to true will attempt to fix these cases by adjusting the 
# time for the 2nd change to be one second more than the original time.
# I dont know why you'd want to turn this off, but I give you the option 
# anyhow. :)
TIME_ADJUSTMENT_HACK = True

###########################################################################
### You probably don't need to change any configuration past this line. ###
###########################################################################

###
### Script begins here
###

import os
import re
import sys
import string
import io

from trac.env import Environment
import trac

# simulated Attachment class for trac.add
# unused in 1.2
class Attachment:
    def __init__(self, name, data):
        self.filename = name
        self.file = io.StringIO(data.tostring())
  
# simple field translation mapping.  if string not in
# mapping, just return string, otherwise return value
class FieldTranslator(dict):
    def __getitem__(self, item):
        if not dict.has_key(self, item):
            return item
            
        return dict.__getitem__(self, item)

class TracDatabase(object):
    def __init__(self, path, append):
        self.append = append
        self.env = Environment(path)
        self._db = self.env.db_transaction
        self._db.autocommit = False
        self.loginNameCache = {}
        self.fieldNameCache = {}

    def hasTickets(self):
        with self.env.db_query as db:
            c = db.cursor()
            c.execute('''SELECT count(*) FROM ticket''')
            return int(c.fetchall()[0][0]) > 0

    def assertNoTickets(self):
        if not self.append or self.hasTickets():
          raise Exception("Will not modify database with existing tickets!")
          return
    
    def setSeverityList(self, s):
        """Remove all severities, set them to `s`"""
        self.assertNoTickets()
        
        c = self.db().cursor()
        c.execute("""DELETE FROM enum WHERE type='severity'""")
        for value, i in s:
            print("inserting severity ", value, " ", i)
            c.execute("""INSERT INTO enum (type, name, value) VALUES (%s, %s, %s)""",
                      ("severity", value, i,))
        self.db().commit()
    
    def setPriorityList(self, s):
        """Remove all priorities, set them to `s`"""
        self.assertNoTickets()
        
        c = self.db().cursor()
        c.execute("""DELETE FROM enum WHERE type='priority'""")
        for value, i in s:
            print("inserting priority ", value, " ", i)
            c.execute("""INSERT INTO enum (type, name, value) VALUES (%s, %s, %s)""",
                      ("priority", value, i,))
        self.db().commit()
    
    def setComponentList(self, l, key):
        """Remove all components, set them to `l`"""
        self.assertNoTickets()
        
        c = self.db().cursor()
        c.execute("""DELETE FROM component""")
        for comp in l:
            print("inserting component '",comp[key],"', owner",  comp['owner'])
            c.execute("""INSERT INTO component (name, owner) VALUES (%s, %s)""",
                      (comp[key], comp['owner'],))
        self.db().commit()
    
    def setVersionList(self, v, key):
        """Remove all versions, set them to `v`"""
        self.assertNoTickets()
        
        c = self.db().cursor()
        c.execute("""DELETE FROM version""")
        for vers in v:
            print("inserting version ", vers[key])
            c.execute("""INSERT INTO version (name) VALUES (%s)""",
                      (vers[key],))
        self.db().commit()

    def setMilestoneList(self, m, key):
        """Remove all milestones, set them to `m`"""
        self.assertNoTickets()

        with self.env.db_transaction as db:
            c = db.cursor()
            c.execute("""DELETE FROM milestone""")
        for ms in m:
            print("inserting milestone ", ms[key])
            milestone = trac.ticket.model.Milestone(self.env)
            milestone.name = ms[key]
            milestone.due = datetime.fromisoformat(ms['date_order']) if ms['date_order'] else None
            milestone.completed = datetime.fromisoformat(ms['released']) if ms['released'] else None
            milestone.insert()

    def addTicket(self, id, time, changetime, component,
                  severity, priority, owner, reporter, cc,
                  version, milestone, status, resolution,
                  summary, description, keywords):
        with self.env.db_transaction as db:
            c = db.cursor()

            desc = description
            type = 'defect'
            if component == 'Features' or severity == 'feature':
                type = 'enhancement'
            if desc is not None:
                if PREFORMAT_COMMENTS:
                    desc = '{{{\n%s\n}}}' % desc
                else:
                    desc = desc.replace('```', '{{{', 1)
                    desc = desc.replace('```', '}}}', 1)
            print("inserting ticket %s -- \"%s\"" % (id, summary[0:40].replace("\n", " ")))
            c.execute("""INSERT INTO ticket (type, time, changetime, component,
                                         severity, priority, owner, reporter, cc,
                                         version, milestone, status, resolution,
                                         summary, description, keywords)
                                 VALUES (%s, %s, %s, %s,
                                         %s, %s, %s, %s, %s,
                                         %s, %s, %s, %s,
                                         %s, %s, %s)""",
                  (type, self.convertTime(time), self.convertTime(changetime), component,
                  severity, priority, owner, reporter, cc,
                  version, milestone, status.lower(), resolution,
                  summary, desc, ' '.join(keywords)))

            ## TODO: add database-specific methods to get the last inserted ticket's id...
            ## PostgreSQL:
            # c.execute('''SELECT currval("ticket_id_seq")''')
            ## SQLite:
            # c.execute('''SELECT last_insert_rowid()''')
            ## MySQL:
            # c.execute('''SELECT LAST_INSERT_ID()''')
            # Oh, Trac db abstraction layer already has a function for this...
            return db.get_last_id(c,'ticket')

    def convertTime(self,time2):
        if time2 is None:
            return None
        time2 = datetime.fromisoformat(time2)
        return int(str(int(time.mktime(time2.timetuple()))) + '000000')

    def addTicketComment(self, ticket, time, author, value):
        print(" * adding comment \"%s...\"" % value[0:40])
        comment = value

        if PREFORMAT_COMMENTS:
          comment = '{{{\n%s\n}}}' % comment

        time = self.convertTime(time)

        with self.env.db_transaction as db:
            c = db.cursor()
            #workaround 'unique' ticket_change warnings POTENTIALLY BAD IDEA ALERT
            sql = "SELECT * FROM ticket_change WHERE field='%s' AND ticket=%s AND time=%s" % ('comment', ticket, time)
            c.execute(sql)
            fixtime = c.fetchall()
            if fixtime:
                time = time + 1
            c.execute("""INSERT  INTO ticket_change (ticket, time, author, field, oldvalue, newvalue)
                                 VALUES        (%s, %s, %s, %s, %s, %s)""",
                  (ticket, time, author, 'comment', '', comment))

    def addTicketChange(self, ticket, time, author, field, oldvalue, newvalue):
        if (field[0:4]=='doba'): 
          return

        print(" * adding ticket change \"%s\": \"%s\" -> \"%s\" (%s)" % (field, oldvalue[0:20], newvalue[0:20], time))

        ctime = self.convertTime(time)
        with self.env.db_transaction as db:
            c = db.cursor()

            fixtime = True
            while fixtime:
                #workaround 'unique' ticket_change warnings POTENTIALLY BAD IDEA ALERT
                sql = "SELECT * FROM ticket_change WHERE field='%s' AND ticket=%s AND time=%s" % (field, ticket, ctime)
                c.execute(sql)
                fixtime = c.fetchall()
                if fixtime:
                    ctime = ctime + 1
            c.execute("""INSERT  INTO ticket_change (ticket, time, author, field, oldvalue, newvalue)
                                 VALUES        (%s, %s, %s, %s, %s, %s)""",
                  (ticket, ctime, author, field, oldvalue, newvalue))

    # unused in 1.2
    def addAttachment(self, id, attachment, description, author):
        print('inserting attachment for ticket %s -- %s' % (id, description))
        attachment.filename = attachment.filename.encode('utf-8')
        self.env.create_attachment(self.db(), 'ticket', str(id), attachment, description.encode('utf-8'),
            author, 'unknown')
        
    def getLoginName(self, cursor, userid):
        if userid not in self.loginNameCache and userid is not None and userid != '':
            cursor.execute("SELECT username,email,realname,last_visit FROM mantis_user_table WHERE id = %i" % int(userid))
            result = cursor.fetchall()

            if result:
                loginName = result[0]['username']
                print('Adding user %s to sessions table' % loginName)
                c = self.db().cursor()

                # check if user is already in the sessions table
                c.execute("SELECT sid FROM session WHERE sid = '%s'" % result[0]['username'].encode('utf-8'))
                r = c.fetchall()
                
                # if there was no user sid in the database already
                if not r:
                    sessionSql = """INSERT INTO session 
                            (sid, authenticated, last_visit) 
                        VALUES (%s, %s, %d)""", (result[0]['username'].encode('utf-8'), '1', self.convertTime(result[0]['last_visit']))
                    # pre-populate the session table and the realname/email table with user data
                    try:
                        c.execute(sessionSql)
                    except:
                        print('failed executing sql: ')
                        print(sessionSql)
                        print('could not insert %s into sessions table: sql error ' % (loginName))
                    self.db().commit()
                
                    # insert the user's real name into session attribute table
                    try:
                        c.execute(
                        """INSERT INTO session_attribute 
                            (sid, authenticated, name, value)
                        VALUES
                            (%s, %s, %s, %s)""", (result[0]['username'].encode('utf-8'), '1', 'name', result[0]['realname'].encode('utf-8')))
                    except:
                        print('failed executing session-attribute sql')
                        self.db().commit()

                    # insert the user's email into session attribute table
                    try:
                        c.execute(
                        """INSERT INTO session_attribute 
                            (sid, authenticated, name, value)
                        VALUES
                            (%s, %s, %s, %s)""", (result[0]['username'].encode('utf-8'), '1', 'email', result[0]['email'].encode('utf-8')))
                    except:
                        print('failed executing session-attribute sql2')
                    self.db().commit()
            else:
                print('warning: unknown mantis userid %d, recording as anonymous' % userid)
                loginName = ''

            self.loginNameCache[userid] = loginName
        elif userid is None or userid == '':
            self.loginNameCache[userid] = ''
        return self.loginNameCache[userid]

    def get_attachments_dir(self,bugid=0):
        if bugid > 0:
            return 'importfiles/%i/' % bugid        
        else:
            return 'importfiles/'

    def _mkdir(newdir):
        """works the way a good mkdir should :)
            - already exists, silently complete
            - regular file in the way, raise an exception
            - parent directory(ies) does not exist, make them as well
        """
        if os.path.isdir(newdir):
            pass
        elif os.path.isfile(newdir):
            raise OSError("a file with the same name as the desired " \
                          "dir, '%s', already exists." % newdir)
        else:
            head, tail = os.path.split(newdir)
            if head and not os.path.isdir(head):
                _mkdir(head)
            #print "_mkdir %s" % repr(newdir)
            if tail:
                os.mkdir(newdir)

def productFilter(fieldName, products):
    first = True
    result = ''
    for product in products:
        if not first: 
            result += " or "
        first = False
        result += "%s = '%s'" % (fieldName, product)
    return result

def convert(github_repo, token, _env, _force, _append):
    activityFields = FieldTranslator()

    # init Trac environment
    print("Trac database('%s'): connecting..." % (_env))
    trac = TracDatabase(_env, _append)

    # force mode...
    if _force == 1:
        print("cleaning all tickets...")
        with trac.env.db_transaction as db:
            c = db.cursor()
            sql = """DELETE FROM ticket_change"""
            c.execute(sql)
            sql = """DELETE FROM ticket"""
            c.execute(sql)
            sql = """DELETE FROM ticket_custom"""
            c.execute(sql)
            sql = """DELETE FROM attachment"""
            c.execute(sql)
            os.system('rm -rf %s' % trac.get_attachments_dir())
            os.mkdir(trac.get_attachments_dir())

    #trac.setSeverityList(SEVERITY_LIST)
    #trac.setComponentList(components, 'category')
    #trac.setPriorityList(PRIORITY_LIST)
    #trac.setVersionList(versions, 'version')

    print()
    print('5. retrieving milestones...')
    headers = {"Authorization": "Bearer " + token}
    req = requests.get("https://api.github.com/repos/" + github_repo + "/milestones?state=all&per_page=100&direction=asc", headers = headers)
    milestones = req.json()

    tracms = []

    for milestone in milestones:
        ms = {}
        ms['version'] = milestone['title']
        ms['date_order'] = milestone['due_on']
        ms['released'] = milestone['closed_at']
        tracms.append(ms)
    trac.setMilestoneList(tracms, 'version')

    print()
    print('6. retrieving bugs...')
    req = requests.get("https://api.github.com/repos/" + github_repo + "/issues?state=all&per_page=100&direction=asc", headers = headers)
    bugs = req.json()
    
    print()
    print("7. import bugs and bug activity...")
    totalComments = 0
    totalTicketChanges = 0
    totalAttachments = 0
    errors = []
    timeAdjustmentHacks = []

    for bug in bugs:
        bugid = bug['number']

        ticket = {}
        keywords = []
        ticket['id'] = bugid
        ticket['time'] = bug['created_at']
        ticket['changetime'] = bug['updated_at']
        ticket['component'] = ''
        ticket['severity'] = ''
        ticket['priority'] = ''
        if bug['assignee']:
            ticket['owner'] = bug['assignee']['login']
        else:
            ticket['owner'] = ''
        ticket['reporter'] = bug['user']['login']
        ticket['version'] = ''
        if bug['milestone']:
            ticket['milestone'] = bug['milestone']['title']
        else:
            ticket['milestone'] = ''
        ticket['summary'] = bug['title']
        if bug['state'] == 'closed':
            ticket['status'] = 'closed'
        elif bug['state_reason'] == 'reopened':
            ticket['status'] = 'reopened'
        else:
            ticket['status'] = 'new'
        ticket['cc'] = ''
        ticket['keywords'] = []
        for label in bug['labels']:
            ticket['keywords'].append(label['name'])

        if 'pull_request' in bug:
            ticket['keywords'].append('pull_request')

        ticket['resolution'] = ''

        ticket['description'] = bug['body']

        # Add the ticket to the Trac database
        new_id = trac.addTicket(**ticket)
        print("ticket %s has id %s" % (bugid, new_id))

        if bug['comments'] != 0:
            comments_url = bug['comments_url']
            req = requests.get(comments_url, headers = headers)
            data = req.json()
            for note in data:
                trac.addTicketComment(new_id, note['created_at'], note['user']['login'], note['body'])

        #
        # Convert ticket changes
        #
        events_url = bug['events_url']
        req = requests.get(events_url, headers = headers)
        data = req.json()
        for event in data:
            if event['event'] == 'closed':
                trac.addTicketChange (new_id, event['created_at'], event['actor']['login'], 'status', 'new', 'closed')
            elif event['event'] == 'labeled':
                trac.addTicketChange (new_id, event['created_at'], event['actor']['login'], 'keywords', '', event['label']['name'])
            elif event['event'] == 'milestoned':
                trac.addTicketChange (new_id, event['created_at'], event['actor']['login'], 'milestone', '', event['milestone']['title'])
            elif event['event'] == 'referenced':
                trac.addTicketComment(new_id, event['created_at'], event['actor']['login'], f"This ticket was mentioned in {event['commit_id']}")
            elif event['event'] == 'merged':
                trac.addTicketComment(new_id, event['created_at'], event['actor']['login'], f"This pull request was merged in {event['commit_id']}")
            elif event['event'] == 'renamed':
                trac.addTicketChange (new_id, event['created_at'], event['actor']['login'], 'summary', event['rename']['from'], event['rename']['to'])
            elif event['event'] == 'review_requested':
                trac.addTicketComment(new_id, event['created_at'], event['actor']['login'], f"Requesting review by {event['requested_reviewer']['login']}")
            elif event['event'] == 'base_ref_force_pushed' or event['event'] == 'mentioned' or event['event'] == 'head_ref_deleted' or event['event'] == 'marked_as_duplicate' or event['event'] == 'pinned' or event['event'] == 'head_ref_force_pushed':
                pass # Who cares about that...
            elif event['event'] == 'subscribed':
                trac.addTicketChange (new_id, event['created_at'], event['actor']['login'], 'cc', '', event['actor']['login'])
            elif event['event'] == 'assigned':
                trac.addTicketChange (new_id, event['created_at'], event['actor']['login'], 'owner', '', event['assignee']['login'])
            elif event['event'] == 'reopened':
                trac.addTicketChange (new_id, event['created_at'], event['actor']['login'], 'status', 'closed', 'reopened')
            else:
                print(event)
                exit()

    print()
    if TIME_ADJUSTMENT_HACK:
        for adjustment in timeAdjustmentHacks:
            print(adjustment)
    if len(errors) != 0:
        print("Some errors occurred while importing:")
        for error in errors:
            print(error)
    else: 
        print("Success!")
    print()
    print("Total tickets imported: %d" % len(bugs))
    print("Total ticket comments:  %d" % totalComments)
    print("Total ticket changes:   %d" % totalTicketChanges)
    print("Total attachments:      %d" % totalAttachments)
    print()

def usage():
    print("github2trac - Imports a bug database from GitHub into Trac.")
    print()
    print("Usage: github2trac.py [options]")
    print()
    print("Available Options:")
    print("  --repo owner/repo             - GitHub repository to import (as used in URL)")
    print("  --token token_string          - GitHub authentication token")
    print("  --tracenv /path/to/trac/env/  - Full path to Trac environment")
    print("  -c | --clean                  - Remove current Trac tickets before importing")
    print("  -a | --append                 - Append bugs to existing project")
    print("  --help | help                 - This help info")
    print()
    print("Additional configuration options can be defined directly in the script.")
    print()
    sys.exit(0)

def main():
    global GITHUB_TOKEN, GITHUB_REPO, TRAC_CLEAN, TRAC_APPEND
    if len (sys.argv) > 1:
        if sys.argv[1] in ['--help','help'] or len(sys.argv) < 1:
            usage()
        iter = 1
        while iter < len(sys.argv):
            if sys.argv[iter] in ['--token'] and iter+1 < len(sys.argv):
                GITHUB_TOKEN = sys.argv[iter+1]
                iter = iter + 1
            elif sys.argv[iter] in ['--repo'] and iter+1 < len(sys.argv):
                GITHUB_REPO = sys.argv[iter+1]
                iter = iter + 1
            elif sys.argv[iter] in ['--tracenv'] and iter+1 < len(sys.argv):
                TRAC_ENV = sys.argv[iter+1]
                iter = iter + 1
            elif sys.argv[iter] in ['-c', '--clean']:
                TRAC_CLEAN = 1
            elif sys.argv[iter] in ['-a', '--append']:
                TRAC_APPEND = 1
            else:
                print("Error: unknown parameter: " + sys.argv[iter])
                sys.exit(0)
            iter = iter + 1
    else:
        usage()

    convert(GITHUB_REPO, GITHUB_TOKEN, TRAC_ENV, TRAC_CLEAN, TRAC_APPEND)

if __name__ == '__main__':
    main()
